"""


"""
from tropicpy.attacks.kotov_ushakov_simple import *
from tropicpy.attacks.kotov_ushakov import *
from tropicpy.protocols.grigorie_shpilrain_2019 import *
from tropicpy.attacks.rudy_monico import *
from tropicpy.attacks.isaac_kahrobaei import *
from tropicpy.protocols.grigorie_shpilrain_2014 import *


def tropical_demo():
    print("In tropical algebra we have:")

    print("0 = " + str(tropical_0))
    print("1 = " + str(tropical_1))

    print("0(3x3) \n=\n" + str(tropical_matrix_0(3)))
    print("1(3x3) \n=\n" + str(tropical_matrix_1(3)))

    print("Examples of tropical operations:")

    a = generate_random_tropical_value(-100, 100)
    b = generate_random_tropical_value(-100, 100)

    print("a) addition:")
    print("(" + str(a) + ")+(" + str(b) + ")=" + str(a + b))

    print("b) multiplication:")
    print("(" + str(a) + ")+(" + str(b) + ")=" + str(a * b))

    print("Examples of operations on tropical matrices:")

    e = generate_random_tropical_matrix(3, -100, 100)
    f = generate_random_tropical_matrix(3, -100, 100)

    print("a) addition:")
    print(e)
    print("+")
    print(f)
    print("=")
    print(e + f)

    print("b) multiplication:")

    print(e)
    print("*")
    print(f)
    print("=")
    print(e * f)

    print("c) power")
    print(e)
    print("** 4\n=")
    print(e ** 4)

    print("d) adjoint multiplication:")

    print(e)
    print("@")
    print(f)
    print("=")
    print(e @ f)

    print("e) adjoint power")
    print(e)
    print("^3\n=")
    print(e ^ 3)


def protocol_demo(protocol):
    print("Example of " + protocol.__name__ + " protocol:")

    n = 3

    print("Parameters:")

    Alice = None
    Bob = None

    if protocol == GrigorieShpilrain2014:
        A = generate_random_tropical_matrix(n, -10 ** 10, 10 ** 10)
        B = generate_random_tropical_matrix(n, -10 ** 10, 10 ** 10)

        print("A = \n" + str(A))
        print("B = \n" + str(B))

        g = random.randint(1, 10)
        print("g = " + str(g))

        Alice = protocol(A, B, g, -1000, 1000)
        Bob = protocol(A, B, g, -1000, 1000)

        print("n = " + str(Alice.n))
    elif protocol == GrigorieShpilrain2019:
        A = generate_random_tropical_matrix(n, -1000, 1000)
        B = generate_random_tropical_matrix(n, -1000, 1000)

        print("A = \n" + str(A))
        print("B = \n" + str(B))

        Alice = protocol(A, B)
        Bob = protocol(A, B)

        print("k = " + str(Alice.k))

    U = Alice.send_message()
    V = Bob.send_message()

    print("Alice's message: \n" + str(U))
    print("Bob's message: \n" + str(V))

    Alice.set_Key(V)
    Bob.set_Key(U)

    if Alice.get_Key() == Bob.get_Key():
        print("Alice and Bob share a secret!")
    else:
        print("Something went wrong!")


def attack_demo(protocol, attack):
    print("Example of " + attack.__name__ + " attack:")

    n = 3

    print("Parameters:")

    Alice = None
    Bob = None

    if protocol == GrigorieShpilrain2014:
        A = generate_random_tropical_matrix(n, -10 ** 10, 10 ** 10)
        B = generate_random_tropical_matrix(n, -10 ** 10, 10 ** 10)

        print("A = \n" + str(A))
        print("B = \n" + str(B))

        g = random.randint(1, 10)
        print("g = " + str(g))

        Alice = protocol(A, B, g, -1000, 1000)
        Bob = protocol(A, B, g, -1000, 1000)

        print("n = " + str(Alice.n))
    elif protocol == GrigorieShpilrain2019:
        A = generate_random_tropical_matrix(n, -1000, 1000)
        B = generate_random_tropical_matrix(n, -1000, 1000)

        print("A = \n" + str(A))
        print("B = \n" + str(B))

        Alice = protocol(A, B)
        Bob = protocol(A, B)

        print("k = " + str(Alice.k))

    U = Alice.send_message()
    V = Bob.send_message()

    print("U:\n" + str(U))
    print("V:\n" + str(V))

    Alice.set_Key(V)

    attack_K = None

    if protocol == GrigorieShpilrain2014:
        if attack == kotov_ushakov_simple:
            attack_K = attack(A, B, g, U, V)
        elif attack == kotov_ushakov:
            attack_K = attack(A, B, g, U, V, -1000)
    elif protocol == GrigorieShpilrain2019:
        attack_K = attack(A, B, U, V)

    if Alice.check_Key(attack_K):
        print("Key was found!")
        print("K = \n" + str(attack_K))
    else:
        print("Keys don't match..")
        print("Protocol key:\n" + str(Alice.get_Key()))
        print("Attack key:\n" + str(attack_K))


def tropicpy_demo():
    protocol = None
    attack = None

    print("What do you want to see?")
    print("1. Tropical algebra examples")
    print("2. Grigorie-Shpilrain (2014) protocol")
    print("3. Grigorie-Shpilrain (2019) protocol")

    ans1 = input()

    try:
        ans1 = int(ans1)

        if ans1 == 1:
            tropical_demo()
        elif ans1 == 2:
            protocol = GrigorieShpilrain2014

            print("What attack do you want to see?")
            print("1. Kotov-Ushakov Simple")
            print("2. Kotov-Ushakov")
            print("3. None, just the protocol")

            ans2 = input()

            try:
                ans2 = int(ans2)

                if ans2 == 1:
                    attack = kotov_ushakov_simple
                elif ans2 == 2:
                    attack = kotov_ushakov
                else:
                    attack = None
            except:
                print("Not a number.")

        elif ans1 == 3:
            protocol = GrigorieShpilrain2019

            print("What attack do you want to see?")
            print("1. Rudy-Monico")
            print("2. Isaac-Kahrobaei")
            print("3. None, just the protocol")

            ans3 = input()
            try:
                ans3 = int(ans3)

                if ans3 == 1:
                    attack = rudy_monico
                elif ans3 == 2:
                    attack = isaac_kahrobaei
                else:
                    attack = None
            except:
                print("Not a number.")

        else:
            print("Nothing to show.")
            return 0
    except:
        print("Not a number.")

    if attack is not None and protocol is not None:
        attack_demo(protocol, attack)
    elif protocol is not None:
        protocol_demo(protocol)


if __name__ == "__main__":
    tropicpy_demo()



"""


"""
from tropicpy.tropical.tropical_matrix import *


def value_difference(a, b):
    if b == tropical_0:
        return False
    elif a == tropical_0:
        return tropical_0
    else:
        return TropicalValue(a.value - b.value)


def matrix_difference(A, B):
    result = []
    for (row_a, row_b) in zip(A.values, B.values):
        tmp_row = []
        for (a, b) in zip(row_a, row_b):
            tmp_row.append(value_difference(a, b))
        result.append(tmp_row)

    return TropicalMatrix(result)


def double_matrix_sum(A, B):
    result = []
    for (row_a, row_b) in zip(A.values, B.values):
        tmp_row = []
        for (a, b) in zip(row_a, row_b):
            tmp_row.append(a * b)
        result.append(tmp_row)

    return TropicalMatrix(result)


def matrix_sum(matrix_table, n=None):
    if not matrix_table:
        return TropicalMatrix([[0] * n] * n)
    R = matrix_table[-1]
    matrix_table.pop()
    for M in matrix_table:
        R = double_matrix_sum(R, M)
    return R


def is_matrix_difference_constant(A, B):
    frst = True
    for (row_a, row_b) in zip(A.values, B.values):
        for (a, b) in zip(row_a, row_b):
            if frst:
                frst = False
                c = value_difference(a, b)
                if not c:
                    continue
            else:
                new_c = value_difference(a, b)
                if new_c:
                    if new_c != c:
                        return False
                    else:
                        c = new_c
    return c


def min_of_matrix_difference(A, B):
    matrix_min = tropical_0
    inds = []
    for (row_a, row_b, i) in zip(A.values, B.values, range(A.rows)):
        for (a, b, j) in zip(row_a, row_b, range(A.columns)):
            c = value_difference(a, b)
            if matrix_min == c:
                inds.append((i, j))
            elif matrix_min + c == c:
                matrix_min += c
                inds = [(i, j)]
    return matrix_min, inds


def is_cover(main_set, cover):
    tmp_main_set = main_set.copy()
    for curr_set in cover:
        for el in curr_set:
            tmp_main_set.remove(el)
    return len(tmp_main_set) == 0


def compress_data(data):
    result = []

    for record in data:
        if record not in result:
            result.append(record)

    return result


def list_union(list_of_lists):
    return list(set(sum(list_of_lists, [])))


def list_concatenation(list_of_lists):
    concatenation = []

    for el in list_of_lists:
        if len(el) > 0:
            concatenation += el

    return concatenation


def list_difference(list_of_lists1, list_of_els2):
    difference = []

    for el in list_of_lists1:
        el_differece = [ind for ind in el if ind not in list_of_els2]
        if len(el_differece) > 0:
            difference.append(el_differece)

    return difference


def list_cartesian(list_of_lists):
    n = len(list_of_lists)

    if n == 1:
        return list_of_lists
    if n == 2:
        tmp_cartesian = []
        for el1 in list_of_lists[0]:
            tmp_cartesian.extend([[el1, el2] for el2 in list_of_lists[1]])
        return tmp_cartesian
    else:
        tmp_cartesian = []
        for el1 in list_of_lists[0]:
            tmp_cartesian.extend(list_concatenation([[el1], el2]) for el2 in list_cartesian(list_of_lists[1:]))
        return tmp_cartesian


def minimal_set_covers(data):
    if len(data) == 0:
        return [[]]

    minimal_cover = []

    sets = compress_data(data)

    for curr_set in sets:
        other_sets_union = list_union([other_cover for other_cover in sets if other_cover != curr_set])
        l = [ind for ind in curr_set if ind not in other_sets_union]
        if len([ind for ind in curr_set if ind not in other_sets_union]) != 0:
            minimal_cover.append(curr_set)

    minimal_cover_union = list_union(minimal_cover)

    uncovered_elements = list_difference(sets, minimal_cover_union)

    if len(uncovered_elements) > 0:
        uncovered_elements.sort(reverse=True, key=(lambda x: len(x)))
        L = minimal_set_covers(list_difference(uncovered_elements, uncovered_elements[0]))
        I = list_concatenation([[[uncovered_elements[0], l] for l in L], [minimal_set_covers(uncovered_elements[1:])]])
        return [list_concatenation([i, minimal_cover]) for i in I]

    return [minimal_cover]


def make_simplex_matrix(data, g, ij):
    matrix = [[0 for i in range(2 * (g + 1) + (g + 1) ** 2 + 1)] for j in range((g + 1) ** 2 + 1)]

    for i in range(g + 1):
        for j in range(g + 1):
            matrix[i * (g + 1) + j][i] = 1
            matrix[i * (g + 1) + j][g + 1 + j] = 1

    for record in data:
        matrix[record["i"] * (g + 1) + record["j"]][2 * (g + 1) + (g + 1) ** 2] = record["minval"].value * (-1)

    for i in range((g + 1) ** 2):
        matrix[i][2 * (g + 1) + i] = -1

    for inds in ij:
        matrix[inds[0] * (g + 1) + inds[1]][2 * (g + 1) + inds[0] * (g + 1) + inds[1]] = 0

    for i in range(len(matrix[0])):
        tmp_sum = 0
        for j in range(len(matrix) - 1):
            tmp_sum += matrix[j][i]
        matrix[(g + 1) ** 2][i] = tmp_sum * (-1)

    return matrix


def find_pivot_column(matrix, matrix_rows, matrix_cols):
    best = 0
    best_i = False
    for i in range(matrix_cols - 1):
        t = matrix[matrix_rows - 1][i]
        if t < best:
            best = t
            best_i = i
    return best_i


def find_pivot_row(matrix, column, matrix_rows, matrix_cols):
    best_i = False
    best = float(inf)
    for i in range(matrix_rows - 1):
        a = matrix[i][column]
        b = matrix[i][matrix_cols - 1]
        if a > 0 and b / a < best:
            best_i = i
            best = b / a
    return best_i


def recalc(simplex_matrix, pivot_row, pivot_col, bs, ns, simplex_matrix_rows, simplex_matrix_cols):
    tmp = ns[pivot_col]
    ns[pivot_col] = bs[pivot_row]
    bs[pivot_row] = tmp

    for i in range(simplex_matrix_rows):
        for j in range(simplex_matrix_cols):
            if i != pivot_row and j != pivot_col:
                simplex_matrix[i][j] = simplex_matrix[i][j] - simplex_matrix[i][pivot_col] * simplex_matrix[pivot_row][
                    j] / simplex_matrix[pivot_row][pivot_col]

    for i in range(simplex_matrix_rows):
        if i != pivot_row:
            simplex_matrix[i][pivot_col] = -simplex_matrix[i][pivot_col] / simplex_matrix[pivot_row][pivot_col]

    for j in range(simplex_matrix_cols):
        if j != pivot_col:
            simplex_matrix[pivot_row][j] = simplex_matrix[pivot_row][j] / simplex_matrix[pivot_row][pivot_col]

    simplex_matrix[pivot_row][pivot_col] = 1 / simplex_matrix[pivot_row][pivot_col]


def simplex(simplex_matrix):
    simplex_matrix_rows = len(simplex_matrix)
    simplex_matrix_cols = len(simplex_matrix[0])
    bs = list(range(simplex_matrix_cols - 1, simplex_matrix_cols + simplex_matrix_rows - 2))
    ns = list(range(simplex_matrix_cols - 1))
    while True:
        pivot_col = find_pivot_column(simplex_matrix, simplex_matrix_rows, simplex_matrix_cols)
        if type(pivot_col) == bool and not pivot_col:
            break
        pivot_row = find_pivot_row(simplex_matrix, pivot_col, simplex_matrix_rows, simplex_matrix_cols)
        if type(pivot_col) == bool and not pivot_row:
            return False
        recalc(simplex_matrix, pivot_row, pivot_col, bs, ns, simplex_matrix_rows, simplex_matrix_cols)

    if simplex_matrix[simplex_matrix_rows - 1][simplex_matrix_cols - 1] != 0:
        return False

    result = [0.0] * (max(bs) + 1)

    for i in range(len(bs)):
        result[bs[i]] = simplex_matrix[i][simplex_matrix_cols - 1]
    return result


def is_zero_modulo(A, B):
    for (row_a, row_b) in zip(A.values, B.values):
        for (a, b) in zip(row_a, row_b):
            if a.value % b.value != 0:
                return False
    return True


def is_zero_matrix(A):
    if not isinstance(A, TropicalMatrix):
        return False
    for row_a in A.values:
        for a in row_a:
            if a.value != 0:
                return False
    return True



"""


"""

from tropicpy.attacks.attacks_utils import *


def isaac_kahrobaei(A, B, U, V):
    m = 1

    D = None
    A_now = A

    previous_D = []
    previous_A = [None]

    while semidirect_power_1st(A, B, m) != U:

        additional_previous_D = []

        while D not in additional_previous_D:
            A_before = A_now
            A_now = semidirect_product_1st(A_before, None, A, B)
            additional_previous_D.append(D)
            previous_A.append(A_before)
            D = matrix_difference(A_now, A_before)

        previous_D = previous_D + additional_previous_D

        previous_D.reverse()
        d = len(previous_D) - previous_D.index(D) - 2
        previous_D.reverse()

        rho = len(previous_D) - (d + 1)
        print("d=" + str(d))
        print("rho=" + str(rho))

        if any(is_zero_matrix(prev_D) for prev_D in previous_D) or is_zero_matrix(D):
            m = d + 1
            break

        Y = matrix_difference(U, previous_A[d + 1])
        full_sum = matrix_sum([previous_D[i + 1] for i in range(d, d + rho)])

        for a in range(1, rho + 1):
            partial_sum = matrix_sum([previous_D[i] for i in range(d + 1, d + a + 1)], D.rows)
            diff = matrix_difference(Y, partial_sum)
            if is_zero_modulo(diff, full_sum):
                print("k=" + str(a))
                break

        x = (diff.values[0][0]).value // (full_sum.values[0][0]).value
        print("x=" + str(x))
        m = d + x * rho + a + 1

    print("Attack was succesfull!")
    print("m = " + str(m))
    Bm = semidirect_power_2nd(A, B, m)

    return semidirect_product_1st(V, None, U, Bm)



"""


"""
from tropicpy.tropical.tropical_matrix import *


class GrigorieShpilrain2019:

    def __init__(self, A, B):
        if not isinstance(A, TropicalMatrix):
            raise Exception(str(A) + " is not an appropriate value.")
        elif not isinstance(B, TropicalMatrix):
            raise Exception(str(B) + " is not an appropriate value.")
        else:
            if A.rows != B.rows or A.columns != B.columns:
                raise Exception("Matrices A and B are of different dimensions.")
            elif A.rows != A.columns:
                raise Exception("Matrix A is not a square matrix.")
            elif B.rows != B.columns:
                raise Exception("Matrix B is not a square matrix.")

            self.k = A.rows
            self.A = A
            self.B = B

            self.m = random.getrandbits(int(2 ** 200).bit_length())

            self.U = None
            self.Bm = None
            self._K = None

    def send_message(self):
        self.U = semidirect_power_1st(self.A, self.B, self.m)
        self.Bm = semidirect_power_2nd(self.A, self.B, self.m)
        return self.U

    def set_Key(self, V):
        self._K = (V @ self.Bm) + self.U

    def get_Key(self):
        return self._K

    def check_Key(self, check_K):
        return check_K == self._K



"""

"""

from tropicpy.tropical.tropical_matrix import *


def rudy_monico(A, B, U, V):
    r = 1
    tmp_A = semidirect_power_1st(A, B, r)
    while U <= tmp_A:
        r *= 2
        tmp_A = semidirect_power_1st(A, B, r)
    print("Upper bound found: \n" + str(r))
    upper = r
    lower = upper >> 1
    while upper >= lower:
        middle = (lower + upper) >> 1
        tmp_A = semidirect_power_1st(A, B, middle)
        if tmp_A == U:
            print("Attack was succesful!")
            break
        elif tmp_A <= U:
            upper = middle - 1
        else:
            lower = middle + 1

    print("m=" + str(middle))
    Bm = semidirect_power_2nd(A, B, middle)
    return semidirect_product_1st(V, None, U, Bm)
